//this file is part of eMule
//Copyright (C)2002 Merkur ( merkur-@users.sourceforge.net / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "StdAfx.h"
#include <zlib/zlib.h>
#include "UpDownClient.h"
#include "opcodes.h"
#include "packets.h"
#include "emule.h"
#include "uploadqueue.h"
#include "otherstructs.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif


//	members of CUpDownClient
//	which are mainly used for uploading functions 

CBarShader CUpDownClient::s_UpStatusBar(16);

void CUpDownClient::DrawUpStatusBar(CDC* dc, RECT* rect, bool onlygreyrect, bool  bFlat){ 
	COLORREF crBoth; 
	COLORREF crNeither; 
	COLORREF crClientOnly; 
	COLORREF crSending;
	COLORREF crNextSending;

	if(bFlat) { 
		crBoth = RGB(0, 0, 0);
		crNeither = RGB(224, 224, 224);
		crClientOnly = RGB(0, 220, 255);
		crSending = RGB(0,150,0);
		crNextSending = RGB(255,208,0);
	} else { 
		crBoth = RGB(104, 104, 104);
		crNeither = RGB(224, 224, 224);
		crClientOnly = RGB(0, 220, 255);
		crSending = RGB(0, 150, 0);
		crNextSending = RGB(255,208,0);
	} 

	// wistily: UpStatusFix
	CKnownFile* currequpfile = theApp.sharedfiles->GetFileByID(requpfileid);
	uint32 filesize;
	if (currequpfile)
		filesize=currequpfile->GetFileSize();
	else
		filesize=PARTSIZE*(m_nUpPartCount);
	// wistily: UpStatusFix

	s_UpStatusBar.SetFileSize(filesize); 
	s_UpStatusBar.SetHeight(rect->bottom - rect->top); 
	s_UpStatusBar.SetWidth(rect->right - rect->left); 
	s_UpStatusBar.Fill(crNeither); 
	if (!onlygreyrect && m_abyUpPartStatus) { 
		for (uint32 i = 0;i < m_nUpPartCount;i++)
			if(m_abyUpPartStatus[i])
				s_UpStatusBar.FillRange(PARTSIZE*(i),PARTSIZE*(i+1),crBoth);
	}
	//Pawcio: BC (critical section)
	Requested_Block_Struct* block=NULL;
	CSingleLock BlockQueueLock(&BlockQueueMutex, true);//Pawcio: BC
	if (!m_BlockRequests_queue.IsEmpty())
		block = m_BlockRequests_queue.GetHead();
	if(block){
		uint32 start = block->StartOffset/PARTSIZE;
		s_UpStatusBar.FillRange(start*PARTSIZE, (start+1)*PARTSIZE, crNextSending);
	}
	block=NULL;
	if (!m_DoneBlocks_list.IsEmpty())
		block = m_DoneBlocks_list.GetTail();
	if(block){
		uint32 start = block->StartOffset/PARTSIZE;
		s_UpStatusBar.FillRange(start*PARTSIZE, (start+1)*PARTSIZE, crNextSending);
	}
	POSITION pos=m_DoneBlocks_list.GetHeadPosition();
	while(pos){
		Requested_Block_Struct* block = m_DoneBlocks_list.GetNext(pos);
		s_UpStatusBar.FillRange(block->StartOffset, block->EndOffset, crSending);
	}
	BlockQueueLock.Unlock();
	s_UpStatusBar.Draw(dc, rect->left, rect->top, bFlat); 
} 

void CUpDownClient::SetUploadState(EUploadState news){
	m_nUploadState = news;
	m_dwShieldTime = 0;
	theApp.emuledlg->transferwnd.clientlistctrl.RefreshClient(this);
}

uint32 CUpDownClient::GetScore(bool sysvalue, bool isdownloading, bool onlybasevalue){
	//TODO: complete this (friends, uploadspeed, emuleuser etc etc)
	if (!m_pszUsername)
		return 0;

	if (credits == 0){
		ASSERT ( false );
		return 0;
	}
	CKnownFile* currequpfile = theApp.sharedfiles->GetFileByID(requpfileid);
	if(!currequpfile)
		return 0;
	
	// bad clients (see note in function)
	if (credits->GetCurrentIdentState(GetIP()) == IS_IDBADGUY)
		return 0;
	// friend slot
	//MORPH START - Added by IceCream, Anti-leecher feature
	if (IsLeecher())
		return 0;
	//MORPH END   - Added by IceCream, Anti-leecher feature
	if (IsFriend() && GetFriendSlot() && !HasLowID())
		return 0x2FFFFFFF;

	if (IsBanned())
		return 0;

	if (sysvalue && HasLowID() && !(socket && socket->IsConnected())){
		return 0;
	}

	// TODO coded by tecxx & herbert, one yet unsolved problem here:
	// sometimes a client asks for 2 files and there is no way to decide, which file the 
	// client finally gets. so it could happen that he is queued first because of a 
	// high prio file, but then asks for something completely different.
	float fBaseValue;
	if (theApp.glob_prefs->IsPawcioCredits() && theApp.glob_prefs->UseCreditSystem()){
		float filepriority = 1.0f; // standard
		if(GetUploadFileID() != NULL){ 
			if(theApp.sharedfiles->GetFileByID(GetUploadFileID()) != NULL){ 
				switch(theApp.sharedfiles->GetFileByID(GetUploadFileID())->GetUpPriority()){ 
                                //Priority changed for new Credit System (which has ratio from 1 to 100 - instead of 1 - 10)
				case PR_VERYHIGH:
					filepriority = 25.0f;
					break;
				case PR_HIGH: 
					filepriority = 10.0f;					 
					break; 
				case PR_LOW: 
					filepriority = 0.5f;//New
					break; 
				case PR_VERYLOW:
					filepriority = 0.2f;
					break;
				case PR_NORMAL: 
				default: 
					filepriority = 1.0f;//New
					break; 
				} 
			} 
		} 
		// calculate score, based on waitingtime and other factors
		if (onlybasevalue)
			fBaseValue = 1.0f;
		else if (!isdownloading)
			fBaseValue = (float)(::GetTickCount()-GetWaitStartTime())/1000.0f;
		else{
			// 15 minutes bonus or MB bonus (when less opened slots)
			fBaseValue = (float)(m_dwShieldTime + m_dwUploadTime-GetWaitStartTime())/1000.0f;
			ASSERT ( m_dwShieldTime + m_dwUploadTime-GetWaitStartTime() >= 0 ); //oct 28, 02: changed this from "> 0" to ">= 0"
		}
		//	if (HasLowID())
		//		fBaseValue *= 0.8f;
		if (GetDownloadState() == DS_DOWNLOADING)
			fBaseValue *= 1.5f; // since up/down credits are in, this isn't that improtant anymore
		/*if (isfriend)
		fBaseValue *= 2.0f;*/
		ASSERT(credits != NULL);
		//Changed by Pawcio
		float fModifValue = credits->GetScoreRatio(this->GetIP());
		fModifValue *=filepriority;
		fBaseValue *= fModifValue;
		if (!isdownloading && !onlybasevalue){
			if (sysvalue && HasLowID() && !(socket && socket->IsConnected()) ){
				if (!theApp.serverconnect->IsConnected() || theApp.serverconnect->IsLowID() || theApp.listensocket->TooManySockets())
					return 0;
			}
		}
		//Kick irregular Clients Donkeys with XS LSD
	        if ( (this->GetVersion()>589) && ((this->GetSourceExchangeVersion()>0) || (this->GetExtendedRequestsVersion()>0)) && (this->GetClientSoft()==51))
			return 0;

	        if ( (this->GetClientSoft() == SO_EDONKEYHYBRID) &&  ( (this->GetSourceExchangeVersion()>0) || (this->GetExtendedRequestsVersion()>0))) //Fake Hybrids
			return 0;
		if( (IsEmuleClient() || this->GetClientSoft() < 10) && m_byEmuleVersion <= 0x19 )
			fBaseValue *= 0.5f;
		//End by Pawcio

	}
	else {
	int filepriority = 10; // standard
	if(GetUploadFileID() != NULL){ 
		if(theApp.sharedfiles->GetFileByID(GetUploadFileID()) != NULL){ 
			switch(theApp.sharedfiles->GetFileByID(GetUploadFileID())->GetUpPriority()){ 
				case PR_VERYHIGH:
					filepriority = 18;
					break;
				case PR_HIGH: 
					filepriority = 9; 
					break; 
				case PR_LOW: 
					filepriority = 6; 
					break; 
				case PR_VERYLOW:
					filepriority = 2;
					break;
				case PR_NORMAL: 
					default: 
					filepriority = 7; 
				break; 
			} 
		} 
	} 
	// calculate score, based on waitingtime and other factors
	if (onlybasevalue)
		fBaseValue = 100;
	else if (!isdownloading)
		fBaseValue = (float)(::GetTickCount()-GetWaitStartTime())/1000.0f;
	else{
		// 15 minutes bonus or MB bonus (when less opened slots)
		fBaseValue = (float)(m_dwShieldTime + m_dwUploadTime-GetWaitStartTime())/1000.0f;
		ASSERT ( m_dwShieldTime + m_dwUploadTime-GetWaitStartTime() >= 0 ); //oct 28, 02: changed this from "> 0" to ">= 0"
	}
	if(theApp.glob_prefs->UseCreditSystem())
		fBaseValue *= credits->GetScoreRatio(GetIP());
	if (!onlybasevalue)
		fBaseValue *= (float(filepriority)/10.0f); 
	if (!isdownloading && !onlybasevalue){
		if (sysvalue && HasLowID() && !(socket && socket->IsConnected()) ){
			if (!theApp.serverconnect->IsConnected() || theApp.serverconnect->IsLowID() || theApp.listensocket->TooManySockets())
				return 0;
		}
	}
	if( (IsEmuleClient() || this->GetClientSoft() < 10) && m_byEmuleVersion <= 0x19 )
		fBaseValue *= 0.5f;
	}
	return (uint32)fBaseValue+(onlybasevalue ? 0 : 1 );
}

//MORPH START - Added by Yun.SF3, ZZ Upload System
/**
 * Checks if the file this client has requested the power shared file.
 *
 * @return true if the requested file is power shared
 */
bool CUpDownClient::GetPowerShared() {
	if(GetUploadFileID() != NULL &&
       theApp.sharedfiles->GetFileByID(GetUploadFileID()) != NULL) {
		return theApp.sharedfiles->GetFileByID(GetUploadFileID())->GetPowerShared();
	} else {
		return false;
	}
}
//MORPH END - Added by Yun.SF3, ZZ Upload System

// Checks if it is next requested block from another chunk of the actual file or from another file 
// 
// [Returns] 
//   true : Next requested block is from another different chunk or file than last downloaded block 
//   false: Next requested block is from same chunk that last downloaded block 
bool CUpDownClient::IsDifferentPartBlock()
{ 
	Requested_Block_Struct* lastBlock;
	Requested_Block_Struct* currBlock;
	uint32 lastDone = 0;
	uint32 currRequested = 0;
	
	bool different = false;
	
	try {
		// Check if we have good lists and proceed to check for different chunks
		//Pawcio: BC (critical section)
		CSingleLock BlockQueueLock(&BlockQueueMutex, true);//Pawcio: BC
		bool reqnotempty=!m_BlockRequests_queue.IsEmpty();
		bool donenotempty=!m_DoneBlocks_list.IsEmpty();
		if ( reqnotempty && donenotempty)
		{
			// Calculate corresponding parts to blocks
			lastBlock = (Requested_Block_Struct*)m_DoneBlocks_list.GetTail();
			currBlock = (Requested_Block_Struct*)m_BlockRequests_queue.GetHead(); 
			lastDone = lastBlock->StartOffset / PARTSIZE;
			currRequested = currBlock->StartOffset / PARTSIZE; 
             
			// Test is we are asking same file and same part
			if ( lastDone != currRequested)
			{ 
				different = true;
				//theApp.emuleDlg->QueueDebugLogLine(false, "%s: Upload session ended due to new chunk.", this->GetUserName());
			}
			if (md4cmp(lastBlock->FileID, currBlock->FileID) != 0)
			{ 
				different = true;
				theApp.emuledlg->QueueDebugLogLine(false, "%s: Upload session ended due to different file.", this->GetUserName());
			}
		} 
   	}
   	catch(...)
   	{ 
			theApp.emuledlg->QueueDebugLogLine(false, "%s: Upload session ended due to error.", this->GetUserName());
      		different = true; 
   	} 
//	AddDebugLogLine(false, "Debug: User %s, last_done_part (%u) %s (%u) next_requested_part, sent %u Kbs.", GetUserName(), last_done_part, different_part? "!=": "==", next_requested_part, this->GetTransferedUp() / 1024); 

	return different; 
}

class CSyncHelper
{
public:
	CSyncHelper()
	{
		m_pObject = NULL;
	}
	~CSyncHelper()
	{
		if (m_pObject)
			m_pObject->Unlock();
	}
	CSyncObject* m_pObject;
};

bool CUpDownClient::CreateNextBlockPackage(){
	// time critical
	// check if we should kick this client
	// VQB Full Chunk Trans..
	if ( !GetFriendSlot() ){
		// Pawcio: Full Chunks for Release Files
		CKnownFile* file = theApp.sharedfiles->GetFileByID(requpfileid);
		if (theApp.glob_prefs->TransferFullChunks()||
			theApp.glob_prefs->IsFullChunkOnlyForRelease() && file && file->GetUpPriority()==PR_VERYHIGH ||
			file && file->GetPowerShared())
		{
			// VQB to provide full chunk transfers (modified by Tarod)
			if ( theApp.uploadqueue->CheckForTimeOver(this, true) || IsDifferentPartBlock())
			{
				// Pawcio: BC
				toRemove=true;
				addOnWaiting=true;
				// Pawcio: BC
				return false;
			}
		}
		else {
			// Pawcio: Minutes to MB
			// upload minimum specified MB data (which approx. equal 15min. when there is less slots than in original eMule) or 15 min. (no slots minimizing) to this client
			if (GetSessionUp() > MinToMB()){
				if (!m_dwShieldTime){//Shiled time for MB shield (when opened less slots)
					m_dwShieldTime = ::GetTickCount() - m_dwUploadTime;
					if (m_dwShieldTime > 900000) //Max 15 min shield
						m_dwShieldTime = 900000;
				}
				if (theApp.uploadqueue->GetWaitingUserCount()!=0 && theApp.uploadqueue->CheckForTimeOver(this, false)){
					// Pawcio: BC
					toRemove=true;
					addOnWaiting=true;
					// Pawcio: BC
					return false;
				}
			}
		}
	}
	//Pawcio: BC - added critical sections (for separe loading parts thread)
	CFile file;
	byte* filedata = 0;
	CString fullname;
	// -khaos--+++> Statistic to breakdown uploaded data by complete file vs. partfile.
	bool bFromPF = true;
	// <-----khaos-
	CSyncHelper lockFile;
	try{
		// Pawcio: BC
		CSingleLock BlockQueueLock(&BlockQueueMutex, true);//Pawcio: BC
		while (!m_BlockRequests_queue.IsEmpty()){
			Requested_Block_Struct* headblock = m_BlockRequests_queue.GetHead();
			Requested_Block_Struct* currentblock = new Requested_Block_Struct(*headblock);
			BlockQueueLock.Unlock();
			// Pawcio: BC
			CKnownFile* srcfile = theApp.sharedfiles->GetFileByID(currentblock->FileID);
			if (!srcfile){
				delete currentblock;
				throw GetResString(IDS_ERR_REQ_FNF);
			}

			if (srcfile->IsPartFile() && ((CPartFile*)srcfile)->GetStatus() != PS_COMPLETE){
				// Do not access a part file, if it is currently moved into the incoming directory.
				// Because the moving of part file into the incoming directory may take a noticable 
				// amount of time, we can not wait for 'm_FileCompleteMutex' and block the main thread.
				if (!((CPartFile*)srcfile)->m_FileCompleteMutex.Lock(0)){ // just do a quick test of the mutex's state and return if it's locked.
					delete currentblock;
					return false;
				}
				lockFile.m_pObject = &((CPartFile*)srcfile)->m_FileCompleteMutex;
				// If it's a part file which we are uploading the file remains locked until we've read the
				// current block. This way the file completion thread can not (try to) "move" the file into
				// the incoming directory.

				fullname = RemoveFileExtension(((CPartFile*)srcfile)->GetFullName());
			}
			else{
				fullname.Format("%s\\%s",srcfile->GetPath(),srcfile->GetFileName());
			}
		
			uint32 togo;
			if (currentblock->StartOffset > currentblock->EndOffset){
				togo = currentblock->EndOffset + (srcfile->GetFileSize() - currentblock->StartOffset);
			}
			else{
				togo = currentblock->EndOffset - currentblock->StartOffset;
				if (srcfile->IsPartFile() && !((CPartFile*)srcfile)->IsComplete(currentblock->StartOffset,currentblock->EndOffset-1)){
					delete currentblock;
					throw GetResString(IDS_ERR_INCOMPLETEBLOCK);
				}
			}

			if( togo > 184320 ){
				delete currentblock;
				throw GetResString(IDS_ERR_LARGEREQBLOCK);
			}

			if (!srcfile->IsPartFile()){
				// -khaos--+++> This is not a part file...
				bFromPF = false;
				// <-----khaos-
				if (!file.Open(fullname,CFile::modeRead|CFile::osSequentialScan|CFile::shareDenyNone)){
					delete currentblock;
					throw GetResString(IDS_ERR_OPEN);
				}
				file.Seek(currentblock->StartOffset,0);
				
				filedata = new byte[togo+500];
				if (uint32 done = file.Read(filedata,togo) != togo){
					file.SeekToBegin();
					file.Read(filedata + done,togo-done);
				}
				file.Close();
			}
			else{
				CPartFile* partfile = (CPartFile*)srcfile;

				partfile->m_hpartfile.Seek(currentblock->StartOffset,0);
				
				filedata = new byte[togo+500];
				if (uint32 done = partfile->m_hpartfile.Read(filedata,togo) != togo){
					partfile->m_hpartfile.SeekToBegin();
					partfile->m_hpartfile.Read(filedata + done,togo-done);
				}
			}
			if (lockFile.m_pObject){
				lockFile.m_pObject->Unlock(); // Unlock the (part) file as soon as we are done with accessing it.
				lockFile.m_pObject = NULL;
			}

			SetUploadFileID(currentblock->FileID);

			// check extention to decide whether to compress or not
			CString ext=srcfile->GetFileName();ext.MakeLower();
			int pos=ext.ReverseFind('.');
			if (pos>-1) ext=ext.Mid(pos);
			bool compFlag=(ext!=".zip" && ext!=".rar" && ext!=".ace" && ext!=".ogm" );
			if (ext==".avi" && theApp.glob_prefs->GetDontCompressAvi()) compFlag=false;

			// -khaos--+++> We're going to add bFromPF as a parameter to the calls to create packets...
			if (m_byDataCompVer == 1 && compFlag )
				CreatePackedPackets(filedata,togo,currentblock,bFromPF);
			else
				CreateStandartPackets(filedata,togo,currentblock,bFromPF);
			// <-----khaos-
			
			// file statistic
			srcfile->statistic.AddTransferred(currentblock->StartOffset, togo);	// SLUGFILLER: Spreadbars

			// Pawcio: BC
			BlockQueueLock.Lock();
			delete currentblock;
			if (m_BlockRequests_queue.IsEmpty()){
				toRemove=true;
				addOnWaiting=true;
				return false;
			}
			m_DoneBlocks_list.AddHead(m_BlockRequests_queue.RemoveHead());
			delete[] filedata;
			filedata = 0;
		}
		// Pawcio: BC
	}
	catch(CString error){
		OUTPUT_DEBUG_TRACE();
		theApp.emuledlg->QueueDebugLogLine(false,GetResString(IDS_ERR_CLIENTERRORED),GetUserName(),error.GetBuffer());
		toRemoveNow=true;  //Pawcio: BC
		addOnWaiting=false; //Pawcio: BC
		if (filedata)
			delete[] filedata;
		return false;
	}
	catch(CFileException* e){
		TCHAR szError[MAX_CFEXP_ERRORMSG];
		e->GetErrorMessage(szError, ARRSIZE(szError));
		theApp.emuledlg->QueueDebugLogLine(false,_T("Failed to create upload package for %s - %s"),GetUserName(),szError);
		toRemoveNow=true;  //Pawcio: BC
		addOnWaiting=false; //Pawcio: BC
		if (filedata)
			delete[] filedata;
		e->Delete();
		return false;
	}
//	theApp.emuledlg->QueueDebugLogLine(false,"Debug: Packet done. Size: %i",blockpack->GetLength());
	return true;
}

// Pawcio: Minutes to MB
// Change 15 minutes to bytes
uint32 CUpDownClient::MinToMB(){
	float upload=theApp.BC->GetMaxUpload();
	if (upload < 10.0f){
		uint32 result = upload * 10000000 / (UPLOAD_CLIENT_DATARATE / 100);
		if (result < 1000000)
			result = 1000000;
		return result;
	}
	uint32 upPerClient = UPLOAD_CLIENT_DATARATE + theApp.BC->GetCurrDataUpload() * 20; //UPLOAD_CLIENT_DATARATE + datarate/50
	return upPerClient * 900; //900 second ~ 15 minutes
}

void CUpDownClient::ProcessUpFileStatus(char* packet,uint32 size){
	if (m_abyUpPartStatus) {
		delete[] m_abyUpPartStatus;
		m_abyUpPartStatus = NULL;	// added by jicxicmic
	}
	m_nUpPartCount = 0;
	m_nUpCompleteSourcesCount= 0;	// #zegzav:completesrc30b (modify)
	if( size == 16 )
		return;
	CSafeMemFile data((BYTE*)packet,size);
	uchar cfilehash[16];
	data.Read(cfilehash,16);
	CKnownFile* tempreqfile = theApp.sharedfiles->GetFileByID(cfilehash);
	uint16 nED2KUpPartCount;
	data.Read(&nED2KUpPartCount,2);
	if (!nED2KUpPartCount){
		m_nUpPartCount = tempreqfile->GetPartCount();
		m_abyUpPartStatus = new uint8[m_nUpPartCount];
		memset(m_abyUpPartStatus,0,m_nUpPartCount);
	}

	else{
		if (tempreqfile->GetED2KPartCount() != nED2KUpPartCount){
			m_nUpPartCount = 0;
			return;
		}
		m_nUpPartCount = tempreqfile->GetPartCount();
		m_abyUpPartStatus = new uint8[m_nUpPartCount];
		uint16 done = 0;
		while (done != m_nUpPartCount){
			uint8 toread;
			data.Read(&toread,1);
			for (sint32 i = 0;i != 8;i++){
				m_abyUpPartStatus[done] = ((toread>>i)&1)? 1:0;
//				We may want to use this for another feature..
//				if (m_abyUpPartStatus[done] && !tempreqfile->IsComplete(done*PARTSIZE,((done+1)*PARTSIZE)-1))
//					bPartsNeeded = true;
				done++;
				if (done == m_nUpPartCount)
					break;
			}
		}
		if ((GetExtendedRequestsVersion() > 1) && (data.GetLength() - data.GetPosition() > 1))
		{
			uint16 nCount;
			data.Read(&nCount,2);
			SetUpCompleteSourcesCount(nCount);
		}
	}
	tempreqfile->NewAvailPartsInfo();
	theApp.emuledlg->transferwnd.queuelistctrl.RefreshClient(this);
}


// -khaos--+++> Added new parameter: bool bFromPF
void CUpDownClient::CreateStandartPackets(byte* data,uint32 togo, Requested_Block_Struct* currentblock, bool bFromPF){
	// <-----khaos-
	uint32 nPacketSize;
	CMemFile memfile((BYTE*)data,togo);
	if (togo > 10240) 
		nPacketSize = togo/(uint32)(togo/10240);
	else
		nPacketSize = togo;
	while (togo){
		if (togo < nPacketSize*2)
			nPacketSize = togo;
		togo -= nPacketSize;
		// -khaos--+++> Create the packet with the new boolean.
		Packet* packet = new Packet(OP_SENDINGPART,nPacketSize+24, OP_EDONKEYPROT, bFromPF);
		// <-----khaos-
		md4cpy(&packet->pBuffer[0],GetUploadFileID());
		uint32 statpos = (currentblock->EndOffset - togo) - nPacketSize;
		memcpy_amd(&packet->pBuffer[16],&statpos,4);
		uint32 endpos = (currentblock->EndOffset - togo);
		memcpy_amd(&packet->pBuffer[20],&endpos,4);
		memfile.Read(&packet->pBuffer[24],nPacketSize);
		CSingleLock BlockQueueLock(&BlockQueueMutex, true);//Pawcio: BC
		m_BlockSend_queue.AddTail(packet);
	}
}

// -khaos--+++> Added new parameter: bool bFromPF
void CUpDownClient::CreatePackedPackets(byte* data,uint32 togo, Requested_Block_Struct* currentblock, bool bFromPF){
	// <-----khaos-
	BYTE* output = new BYTE[togo+300];
	uLongf newsize = togo+300;
	uint16 result = compress2(output,&newsize,data,togo,9);
	if (result != Z_OK || togo <= newsize){
		delete[] output;
		// -khaos--+++>  Our new boolean...
		CreateStandartPackets(data,togo,currentblock,bFromPF);
		// <-----khaos-
		return;
	}
	m_bUsedComprUp = true;
	CMemFile memfile(output,newsize);
	togo = newsize;
	uint32 nPacketSize;
	if (togo > 10240) 
		nPacketSize = togo/(uint32)(togo/10240);
	else
		nPacketSize = togo;
	while (togo){
		if (togo < nPacketSize*2)
			nPacketSize = togo;
		togo -= nPacketSize;
		// -khaos--+++> Create the packet with the new boolean.
		Packet* packet = new Packet(OP_COMPRESSEDPART,nPacketSize+24,OP_EMULEPROT,bFromPF);
		// <-----khaos-
		md4cpy(&packet->pBuffer[0],GetUploadFileID());
		uint32 statpos = currentblock->StartOffset;
		memcpy_amd(&packet->pBuffer[16],&statpos,4);
		memcpy_amd(&packet->pBuffer[20],&newsize,4);
		memfile.Read(&packet->pBuffer[24],nPacketSize);
	
		CSingleLock BlockQueueLock(&BlockQueueMutex, true);//Pawcio: BC
		m_BlockSend_queue.AddTail(packet);
	}
	delete[] output;
}

void CUpDownClient::SetUploadFileID(uchar* tempreqfileid){
	CKnownFile* newreqfile = NULL;
	if( tempreqfileid )
		newreqfile = theApp.sharedfiles->GetFileByID(tempreqfileid);
	CKnownFile* oldreqfile = theApp.sharedfiles->GetFileByID(requpfileid);
	if(newreqfile == oldreqfile)
		return;
	if(newreqfile){
		newreqfile->AddQueuedCount();
		newreqfile->AddUploadingClient(this);
		md4cpy(requpfileid,tempreqfileid);
	}
	else{
		md4clr(requpfileid);
	}
	if(oldreqfile){
		oldreqfile->SubQueuedCount();
		oldreqfile->RemoveUploadingClient(this);
	}
}


//Xman see OnUploadqueue
void CUpDownClient::SetOldUploadFileID(){
	if( requpfileid )
		md4cpy(oldfileid,requpfileid);
}
//Xman end

void CUpDownClient::AddReqBlock(Requested_Block_Struct* reqblock){
	//Pawcio: BC (critical section)
	CSingleLock BlockQueueLock(&BlockQueueMutex, true);//Pawcio: BC
	for (POSITION pos = m_DoneBlocks_list.GetHeadPosition();pos != 0;m_DoneBlocks_list.GetNext(pos)){
		if (reqblock->StartOffset == m_DoneBlocks_list.GetAt(pos)->StartOffset && reqblock->EndOffset == m_DoneBlocks_list.GetAt(pos)->EndOffset){
			delete reqblock;
			return;
		}
	}
	for (POSITION pos = m_BlockRequests_queue.GetHeadPosition();pos != 0;m_BlockRequests_queue.GetNext(pos)){
		if (reqblock->StartOffset == m_BlockRequests_queue.GetAt(pos)->StartOffset && reqblock->EndOffset == m_BlockRequests_queue.GetAt(pos)->EndOffset){
			delete reqblock;
			return;
		}
	}
	m_BlockRequests_queue.AddTail(reqblock);
}

//Pawcio: BC
uint32	CUpDownClient::GetPacketSizeToSend(){
	CSingleLock BlockQueueLock(&BlockQueueMutex, true);//Pawcio: BC
	if (m_BlockSend_queue.IsEmpty()){
		return 0;
	}
	if(m_BlockSend_queue.GetHead()->IsSplitted() == false && 
		m_BlockSend_queue.GetHead()->GetRealPacketSize() > MAXFRAGSIZE){
			SplitHeadPacket();
		}
	uint32 size=m_BlockSend_queue.GetHead()->GetRealPacketSize();
	return size;
}


// Maella -New bandwidth control- //added by Pawcio: BC
void CUpDownClient::SplitHeadPacket(){

	// Maella -MTU Configuration-
	const uint32 MaxFragSize = MAXFRAGSIZE;
	// Maella end
		
	Packet* pHeadPacket = m_BlockSend_queue.GetHead();
	if(pHeadPacket->IsSplitted() == false && pHeadPacket->GetRealPacketSize() > MaxFragSize){
				// -khaos--+++> When splitting packets, we need to maintain bFromPF, otherwise...
				//				the stats get distorted.
				bool bFromPF = m_BlockSend_queue.GetHead()->IsFromPF();
				// <-----khaos-

		// Split packet
		const uint32 nSize = pHeadPacket->GetRealPacketSize();
		const char* pBuffer = pHeadPacket->DetachPacket();

		// Remove packet from list to split it
				delete m_BlockSend_queue.RemoveHead();
				
		// Average size of splited packet
		const uint32 numberPacket = (nSize + (MaxFragSize - 1)) / MaxFragSize;
		const uint32 sizeSplitedPacket = (nSize + (numberPacket - 1)) / numberPacket; //Pawcio: Proper Split Size

		// Create n packets of the same size (shall improve the ping)
		bool bLast = true;
				uint32 nPos = nSize;
		while(nPos > 0){
			uint32 nNewSize = (nPos < sizeSplitedPacket) ? nPos : sizeSplitedPacket;				
					nPos -= nNewSize;
					char* pBuffer2 = new char[nNewSize];
					memcpy_amd(pBuffer2,pBuffer+nPos,nNewSize);
					// -khaos--+++> Include bFromPF in our new packets to maintain statistical integrity.
					m_BlockSend_queue.AddHead(new Packet(pBuffer2,nNewSize,bLast,bFromPF));
					// <-----khaos-
					bLast = false;
				}

				delete[] pBuffer;
		}
}
// Maella end

// some code from Maella - //Pawcio: BC
uint32 CUpDownClient::SendBlockData(uint32 nMaxAmmount, bool keepAlive){
			
	if (m_AvarageUDR_list.GetCount()>0) {
		// By BadWolf - Accurate speed Measurement
		while ((float)(m_AvarageUDR_list.GetTail().timestamp - m_AvarageUDR_list.GetHead().timestamp)  > MAXAVERAGETIME/2)
			sumavgUDR -= m_AvarageUDR_list.RemoveHead().datalen;
		if (m_AvarageUDR_list.GetCount() > 10 && m_AvarageUDR_list.GetTail().timestamp - m_AvarageUDR_list.GetHead().timestamp > 500)
			m_nUpDatarate = 1000 * sumavgUDR / (m_AvarageUDR_list.GetTail().timestamp - m_AvarageUDR_list.GetHead().timestamp);
		else 
			m_nUpDatarate = 0;
		// END By BadWolf - Accurate speed Measurement
	} else m_nUpDatarate = 0;

	m_cSendblock++;
	if (m_cSendblock == 30){
		m_cSendblock = 0;
		theApp.emuledlg->transferwnd.uploadlistctrl.RefreshClient(this);
		theApp.emuledlg->transferwnd.reservedlistctrl.RefreshClient(this);
		theApp.emuledlg->transferwnd.clientlistctrl.RefreshClient(this);
	}
			
	dataratems = 0;
	// Pawcio: BC
	CSingleLock BlockQueueLock(&BlockQueueMutex, true);//Pawcio: BC
	if (isReady() && nMaxAmmount){
		ASSERT(nMaxAmmount==m_BlockSend_queue.GetHead()->GetRealPacketSize());
		Packet* tosend = m_BlockSend_queue.RemoveHead();
		BlockQueueLock.Unlock();
		dataratems += nMaxAmmount; // By Badwolf
		if (!keepAlive) {
			m_dwLastBlockSent = ::GetTickCount();
			CalcRatio(false);
		}
		m_dwLastBlockKeepAliveSent = ::GetTickCount();
		// -khaos--+++>
		// Extended statistics information based on which client software and which port we sent this data to...
		// This also updates the grand total for sent bytes, etc.  And where this data came from.  Yeesh.
		theApp.glob_prefs->Add2SessionTransferData(GetClientSoft(), GetUserPort(), tosend->IsFromPF(), true, nMaxAmmount);
		socket->SendPacket(tosend,true,false,true);
		m_nTransferedUp += dataratems;
		credits->AddUploaded(dataratems, GetIP());
	}
	TransferredData newitem = {dataratems,::GetTickCount()};
	m_AvarageUDR_list.AddTail(newitem);
	sumavgUDR += dataratems;
	
	return dataratems;
}

void CUpDownClient::FlushSendBlocks(){ // call this when you stop upload, or the socket might be not able to send
	//Pawcio: BC Care of flush taken care in CUploadQueue::Process()
	CSingleLock BlockQueueLock(&BlockQueueMutex, true);//Pawcio: BC
	if (!m_BlockSend_queue.IsEmpty() && m_BlockSend_queue.GetHead()->IsSplitted() && socket && socket->IsConnected() && !socket->HasControlPacket()) 
		socket->ClearLinkedPacket();
}

void CUpDownClient::SendHashsetPacket(char* forfileid){
	CKnownFile* file = theApp.sharedfiles->GetFileByID((uchar*)forfileid);
	if (!file)
		throw GetResString(IDS_ERR_REQ_FNF) + _T(" (SendHashsetPacket)");

	CSafeMemFile data(1024);
	data.Write(file->GetFileHash(),16);
	uint16 parts = file->GetHashCount();
	data.Write(&parts,2);
	for (int i = 0; i < parts; i++)
		data.Write(file->GetPartHash(i),16);
	Packet* packet = new Packet(&data);
	packet->opcode = OP_HASHSETANSWER;
	theApp.uploadqueue->AddUpDataOverheadFileRequest(packet->size);
	socket->SendPacket(packet,true,true);
}

void CUpDownClient::ClearUploadBlockRequests(){
	FlushSendBlocks();
	//Pawcio: BC (critical section)
	CSingleLock BlockQueueLock(&BlockQueueMutex, true);//Pawcio: BC
	for (POSITION pos = m_BlockRequests_queue.GetHeadPosition();pos != 0;m_BlockRequests_queue.GetNext(pos))
		delete m_BlockRequests_queue.GetAt(pos);
	m_BlockRequests_queue.RemoveAll();
	
	for (POSITION pos = m_DoneBlocks_list.GetHeadPosition();pos != 0;m_DoneBlocks_list.GetNext(pos))
		delete m_DoneBlocks_list.GetAt(pos);
	m_DoneBlocks_list.RemoveAll();
	
	for (POSITION pos = m_BlockSend_queue.GetHeadPosition();pos != 0;m_BlockSend_queue.GetNext(pos))
		delete m_BlockSend_queue.GetAt(pos);
	m_BlockSend_queue.RemoveAll();
}

void CUpDownClient::SendRankingInfo(){
	if (!ExtProtocolAvailable())
		return;
	uint16 nRank = theApp.uploadqueue->GetWaitingPosition(this);
	if (!nRank)
		return;
	Packet* packet = new Packet(OP_QUEUERANKING,12,OP_EMULEPROT);
	memset(packet->pBuffer,0,12);
	memcpy_amd(packet->pBuffer+0,&nRank,2);
	// START enkeyDev(th1) -EDT-
	if (GetDownloadTimeVersion() && theApp.glob_prefs->IsEDTEnabled()) {
		uint32 avg_value;
		uint32 err_value;
		EstimateDownloadTime(avg_value, err_value);
		memcpy_amd(packet->pBuffer+4,&avg_value,4);
		memcpy_amd(packet->pBuffer+8,&err_value,4);
	}
	// END enkeyDev(th1) -EDT-
	theApp.uploadqueue->AddUpDataOverheadOther(packet->size);
	socket->SendPacket(packet,true,true);
}

void CUpDownClient::SendCommentInfo(CKnownFile *file) {
	if (!m_bCommentDirty || file == NULL || !ExtProtocolAvailable() || m_byAcceptCommentVer < 1)
		return;
	m_bCommentDirty = false;

	int8 rating=file->GetFileRate();
	CString desc=file->GetFileComment();
	if(file->GetFileRate() == 0 && desc.IsEmpty())
		return;

	CSafeMemFile data(256);
	data.Write(&rating,sizeof(rating));
	int length=desc.GetLength();
	if (length>128)
		length=128;
	data.Write(&length,sizeof(length));
	if (length>0)
		data.Write(desc.GetBuffer(),length);
	Packet *packet = new Packet(&data,OP_EMULEPROT);
	packet->opcode = OP_FILEDESC;
	theApp.uploadqueue->AddUpDataOverheadOther(packet->size);
	socket->SendPacket(packet,true);
}

void  CUpDownClient::AddRequestCount(uchar* fileid){
	for (POSITION pos = m_RequestedFiles_list.GetHeadPosition();pos != 0;m_RequestedFiles_list.GetNext(pos)){
		Requested_File_Struct* cur_struct = m_RequestedFiles_list.GetAt(pos);
		if (!md4cmp(cur_struct->fileid,fileid)){
			if (::GetTickCount() - cur_struct->lastasked < MIN_REQUESTTIME && !GetFriendSlot()){ 
				if (GetDownloadState() != DS_DOWNLOADING)
					cur_struct->badrequests++;
				if (cur_struct->badrequests == BADCLIENTBAN){
					Ban();
				}
			}
			else{
				if (cur_struct->badrequests)
					cur_struct->badrequests--;
			}
			cur_struct->lastasked = ::GetTickCount();
			return;
		}
	}
	Requested_File_Struct* new_struct = new Requested_File_Struct;
	memset(new_struct,0,sizeof(Requested_File_Struct));
	md4cpy(new_struct->fileid,fileid);
	new_struct->lastasked = ::GetTickCount();
	m_RequestedFiles_list.AddHead(new_struct);
}

void  CUpDownClient::UnBan(){
	theApp.clientlist->AddTrackClient(this);
	theApp.clientlist->RemoveBannedClient( GetIP() );
	SetUploadState(US_NONE);
	ClearWaitStartTime(); // Moonlight: SUQWT
	if (credits)
		credits->ClearUploadQueueWaitTime(); // Moonlight: SUQWT
	theApp.emuledlg->transferwnd.ShowQueueCount(theApp.uploadqueue->GetWaitingUserCount());
	for (POSITION pos = m_RequestedFiles_list.GetHeadPosition();pos != 0;m_RequestedFiles_list.GetNext(pos)){
		Requested_File_Struct* cur_struct = m_RequestedFiles_list.GetAt(pos);
		cur_struct->badrequests = 0;
		cur_struct->lastasked = 0;	
	}
//	theApp.emuledlg->transferwnd.queuelistctrl.RefreshClient(this, true, true);
}

void CUpDownClient::Ban(){
	theApp.clientlist->AddTrackClient(this);
	ClearWaitStartTime(); // Moonlight: SUQWT
	if ( !IsBanned() ){
		AddDebugLogLine(false,GetResString(IDS_CLIENTBLOCKED),GetUserName());
	}
#ifdef _DEBUG
	else
		AddDebugLogLine(false,"Ban refreshed for %s (%s) ", GetUserName(), GetFullIP() );
#endif
	theApp.clientlist->AddBannedClient( GetIP() );
	//theApp.uploadqueue->UpdateBanCount();
	SetUploadState(US_BANNED);
	theApp.emuledlg->transferwnd.ShowQueueCount(theApp.uploadqueue->GetWaitingUserCount());
	theApp.emuledlg->transferwnd.queuelistctrl.RefreshClient(this);

}

//MORPH START - Added by IceCream, Anti-leecher feature
void CUpDownClient::BanLeecher(int log_message){
	theApp.clientlist->AddTrackClient(this);
	if (!m_bLeecher){
		theApp.stat_leecherclients++;
		m_bLeecher = true;
	}
	theApp.clientlist->AddBannedClient( GetIP() );
	theApp.emuledlg->transferwnd.ShowQueueCount(theApp.uploadqueue->GetWaitingUserCount());
	theApp.emuledlg->transferwnd.queuelistctrl.RefreshClient(this);
	if (log_message)
		theApp.emuledlg->AddDebugLogLine(false,GetResString(IDS_ANTILEECHERLOG),GetUserName());
}
//MORPH END   - Added by IceCream, Anti-leecher feature


void CUpDownClient::UDPFileReasked(){
	AddAskedCount();
	SetLastUpRequest();
	uint16 nRank = theApp.uploadqueue->GetWaitingPosition(this);
//	Packet* response = new Packet(OP_REASKACK,2,OP_EMULEPROT); // enkeyDev(th1) original, commented
//	memcpy_amd(response->pBuffer,&nRank,2); // enkeyDev(th1) original, commented
	// START enkeyDev(th1) -EDT-
	Packet* response;
	if (GetDownloadTimeVersion() && theApp.glob_prefs->IsEDTEnabled()) {
		response = new Packet(OP_REASKACK,12,OP_EMULEPROT);
		memset(response->pBuffer,0,12);
		memcpy_amd(response->pBuffer,&nRank,2);
		uint32 avg_value;
		uint32 err_value;
		EstimateDownloadTime(avg_value, err_value);
		memcpy_amd(response->pBuffer+4,&avg_value,4);
		memcpy_amd(response->pBuffer+8,&err_value,4);
	}
	else
	{
		response = new Packet(OP_REASKACK,2,OP_EMULEPROT);
		memcpy_amd(response->pBuffer,&nRank,2);
	}
	// END enkeyDev(th1) -EDT-
	theApp.uploadqueue->AddUpDataOverheadFileRequest(response->size);
	theApp.clientudp->SendPacket(response,GetIP(),GetUDPPort());
}

uint32 CUpDownClient::GetWaitStartTime(){
	if (credits == NULL){
		ASSERT ( false );
		return ::GetTickCount() - 1;//Pawcio fix
	}
	uint32 dwResult = credits->GetSecureWaitStartTime(GetIP());
	uint32 curTick = ::GetTickCount(); // Moonlight: SUQWT
	if (curTick - dwResult < curTick - m_dwUploadTime && IsDownloading()){// Moonlight: SUQWT
		//this happens only if two clients with invalid securehash are in the queue - if at all
		dwResult = m_dwUploadTime-1;

		DEBUG_ONLY(AddDebugLogLine(false,"Warning: CUpDownClient::GetWaitStartTime() waittime Collision (%s)",GetUserName()));
	}
	return dwResult;
}

void CUpDownClient::SetWaitStartTime(){
	if (credits == NULL){
		return;
	}
	credits->SetSecWaitStartTime(GetIP());
}

void CUpDownClient::ClearWaitStartTime(){
	if (credits == NULL){
		return;
	}
	credits->ClearWaitStartTime();
}


bool CUpDownClient::GetFriendSlot(){
	if (credits && theApp.clientcredits->CryptoAvailable()){
		switch(credits->GetCurrentIdentState(GetIP())){
			case IS_IDFAILED:
			case IS_IDNEEDED:
			case IS_IDBADGUY:
				return false;
		}
	}
	return m_bFriendSlot;
}//Pawcio: BC
void CUpDownClient::ResetProcessRate(){
	cur_ratio = 0.75f;
	avg_ratio = 0.75f;
	for (int i = 0; i < RATIO_AVG; i++)
		ratios[i] = 0.75f;
	ratioIndex = 0;
	m_uBusyTicks=0;
	m_uReadyTicks=0;
	waitingforparts=false;
	processed=false;
}

void CUpDownClient::CalcRatio(bool busy){
	float multiplier = 0.1f;
	if (busy){
		if (m_uBusyTicks < 40)
			multiplier =log10f(m_uBusyTicks) / 2.4f + 0.33f;
		cur_ratio -= cur_ratio/250 * multiplier;
	}
	else{
		if (m_uReadyTicks < 40)
			multiplier = log10f(m_uReadyTicks) / 2.4f + 0.33f;
		cur_ratio += (1.0f-cur_ratio)/250 * multiplier;
	}
	avg_ratio *= RATIO_AVG;
	avg_ratio -= ratios[ratioIndex];
	avg_ratio += cur_ratio;
	avg_ratio /= RATIO_AVG;
	ratios[ratioIndex++] = cur_ratio;
	ratioIndex %= RATIO_AVG;
	}

//Pawcio: BC

// START enkeyDev(th1) -EDT-
void CUpDownClient::EstimateDownloadTime(uint32 &avg_time, uint32 &err_time)
{
	avg_time = 0;
	err_time = EDT_UNDEFINED;
	if (!IsDownloading())
		theApp.m_edt.EstimateTime(this, avg_time, err_time);
}
// END enkeyDev(th1) -EDT-
